#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# 模型评估脚本

import os
import numpy as np
import json
import paddle
import argparse

import xml.etree.ElementTree as ET

from reader import test_data_loader
from multinms import multiclass_nms
from yolov3 import YOLOv3
from map_utils import DetectionMAP

def parse_args():
    parser = argparse.ArgumentParser("Evaluation Parameters")
    parser.add_argument('--image_dir', type=str, default='./insects/test/images',
        help='the directory of test images')
    parser.add_argument('--weight_files', nargs="+",
        help='the paths of model parameters')
    parser.add_argument('--anno_dir', type=str, default='./insects/test/annotations/xmls',
        help='the directory of annotations')
    parser.add_argument('--dataset_dir', type=str)
    parser.add_argument('--pred_result', type=str, default='./pred_results.json',
        help='the file path of prediction results')
    parser.add_argument('--output_dir', type=str, default='./eval_output',
        help='the dir of best model to save')
    parser.add_argument('--model_names', nargs='+',
        help='names of models, used to get dir of model files save at by env args')
    args = parser.parse_args()
    return args


args = parse_args()
TESTDIR = args.dataset_dir + "/images"

ANCHORS = [10, 13, 16, 30, 33, 23, 30, 61, 62, 45, 59, 119, 116, 90, 156, 198, 373, 326]

ANCHOR_MASKS = [[6, 7, 8], [3, 4, 5], [0, 1, 2]]

VALID_THRESH = 0.01

NMS_TOPK = 400
NMS_POSK = 100
NMS_THRESH = 0.45

NUM_CLASSES = 7

annotations_dir = args.dataset_dir + "/annotations/xmls" # 存放测试集标签的目录
pred_result_file = args.pred_result # 存放预测结果的文件

overlap_thresh = 0.5
map_type = '11point'
insect_names = ['Boerner', 'Leconte', 'Linnaeus', 'acuminatus', 'armandi', 'coleoptera', 'linnaeus']


if __name__ == '__main__':
    res_dict = {}

    # 通过环境变量获取Artifact的存储路径
    # os.getenv(name)的返回值是一个代表路径的字符串，如"./pretrained"
    # 具体讲解可参考：https://cloud.baidu-int.com/icloud/PaddleCloud/%E6%93%8D%E4%BD%9C%E6%8C%87%E5%8D%97/pipeline%E6%93%8D%E4%BD%9C%E6%8C%87%E5%8D%97/%E7%BC%96%E6%8E%92/%E8%AF%A6%E8%A7%A3#%E9%80%9A%E8%BF%87%E7%8E%AF%E5%A2%83%E5%8F%98%E9%87%8F%E8%8E%B7%E5%8F%96
    weight_files = []
    for name in args.model_names:
        weight_files.append(os.getenv(name)+'/param.pdparams')
    
    for params_file_path in weight_files:
        print("Start evaluating model with params file at: " + params_file_path)
        model = YOLOv3(num_classes=NUM_CLASSES)
        model_state_dict = paddle.load(params_file_path)
        model.load_dict(model_state_dict)
        model.eval()

        total_results = []
        test_loader = test_data_loader(TESTDIR, batch_size= 1, mode='test')
        for i, data in enumerate(test_loader()):
            img_name, img_data, img_scale_data = data
            img = paddle.to_tensor(img_data)
            img_scale = paddle.to_tensor(img_scale_data)

            outputs = model.forward(img)
            bboxes, scores = model.get_pred(outputs,
                                    im_shape=img_scale,
                                    anchors=ANCHORS,
                                    anchor_masks=ANCHOR_MASKS,
                                    valid_thresh = VALID_THRESH)

            bboxes_data = bboxes.numpy()
            scores_data = scores.numpy()
            result = multiclass_nms(bboxes_data, scores_data,
                        score_thresh=VALID_THRESH, 
                        nms_thresh=NMS_THRESH, 
                        pre_nms_topk=NMS_TOPK, 
                        pos_nms_topk=NMS_POSK)
            for j in range(len(result)):
                result_j = result[j]
                img_name_j = img_name[j]
                total_results.append([img_name_j, result_j.tolist()])
            if i % 10 == 0 and i != 0:
                print('processed {} pictures'.format(i + 1), flush=True)

        #json.dump(total_results, open('pred_results.json', 'w'))
        

        # 计算mAP
        cname2cid = {}
        for i, item in enumerate(insect_names):
            cname2cid[item] = i
        #filename = pred_result_file
        #results = json.load(open(filename))
        num_classes = len(insect_names)
        detection_map = DetectionMAP(class_num=num_classes,
                            overlap_thresh=overlap_thresh,
                            map_type=map_type,
                            is_bbox_normalized=False,
                            evaluate_difficult=False)
        for result in total_results:
            image_name = str(result[0])
            bboxes = np.array(result[1]).astype('float32')
            anno_file = os.path.join(annotations_dir, image_name + '.xml')
            tree = ET.parse(anno_file)
            objs = tree.findall('object')
            im_w = float(tree.find('size').find('width').text)
            im_h = float(tree.find('size').find('height').text)
            gt_bbox = np.zeros((len(objs), 4), dtype=np.float32)
            gt_class = np.zeros((len(objs), 1), dtype=np.int32)
            difficult = np.zeros((len(objs), 1), dtype=np.int32)
            for i, obj in enumerate(objs):
                cname = obj.find('name').text
                gt_class[i][0] = cname2cid[cname]
                _difficult = int(obj.find('difficult').text)
                x1 = float(obj.find('bndbox').find('xmin').text)
                y1 = float(obj.find('bndbox').find('ymin').text)
                x2 = float(obj.find('bndbox').find('xmax').text)
                y2 = float(obj.find('bndbox').find('ymax').text)
                x1 = max(0, x1)
                y1 = max(0, y1)
                x2 = min(im_w - 1, x2)
                y2 = min(im_h - 1, y2)
                gt_bbox[i] = [x1, y1, x2, y2]
                difficult[i][0] = _difficult
            detection_map.update(bboxes, gt_bbox, gt_class, difficult)

        print("Accumulating evaluatation results fo model with param at " + params_file_path + "......")
        detection_map.accumulate()
        map_stat = 100. * detection_map.get_map()
        res_dict[params_file_path] = map_stat
        print("mAP({:.2f}, {}) = {:.2f}".format(overlap_thresh,
                                map_type, map_stat))
    
    best_model_path = ""
    best_map = 0
    # 保存mAP最大的模型
    for path, res in res_dict.items():
        if best_map < res:
            best_model_path = path
            best_map = res

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)
    paddle.save(model.state_dict(), args.output_dir + '/param.pdparams')
        

