#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# 此文件包括模型训练脚本

import time
import os
import numpy as np
import paddle
import argparse

from reader import TrainDataset
from yolov3 import YOLOv3

# train.py
# 提升点： 可以改变anchor的大小，注意训练和测试时要使用同样的anchor
ANCHORS = [10, 13, 16, 30, 33, 23, 30, 61, 62, 45, 59, 119, 116, 90, 156, 198, 373, 326]

ANCHOR_MASKS = [[6, 7, 8], [3, 4, 5], [0, 1, 2]]

IGNORE_THRESH = .7
NUM_CLASSES = 7

def get_lr(base_lr = 0.0001, lr_decay = 0.1):
    bd = [10000, 20000]
    lr = [base_lr, base_lr * lr_decay, base_lr * lr_decay * lr_decay]
    learning_rate = paddle.optimizer.lr.PiecewiseDecay(boundaries=bd, values=lr)
    return learning_rate


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument("--output_dir", default='./train_output_model', type=str, help="The output directory where the model checkpoints will be written.")
    parser.add_argument("--batch_size", default=10, type=int, help="Batch size per GPU/CPU for training.")
    parser.add_argument("--epoch", default=1, type=int, help="Total number of training epochs to perform.")
    parser.add_argument('--device', default="gpu", help="Select which device to train model, defaults to gpu.")
    parser.add_argument("--ds_dir", type=str, default="train_list.txt")
    parser.add_argument("--param_path", type=str, default="")
    args = parser.parse_args()

    TRAINDIR = args.ds_dir
    paddle.set_device(args.device)
    # 创建数据读取类
    train_dataset = TrainDataset(TRAINDIR, mode='train')
    # 使用paddle.io.DataLoader创建数据读取器，并设置batch size，进程数量num_workers等参数
    train_loader = paddle.io.DataLoader(train_dataset, batch_size=args.batch_size, shuffle=True, num_workers=0, drop_last=True, use_shared_memory=False)
    model = YOLOv3(num_classes = NUM_CLASSES)  #创建模型

    if args.param_path:
        print("Training with params at: " + args.param_path)
        state_dict = paddle.load(args.param_path)
        model.load_dict(state_dict)

    learning_rate = get_lr()
    opt = paddle.optimizer.Momentum(
                 learning_rate=learning_rate,
                 momentum=0.9,
                 weight_decay=paddle.regularizer.L2Decay(0.0005),
                 parameters=model.parameters())  #创建优化器
    # opt = paddle.optimizer.Adam(learning_rate=learning_rate, weight_decay=paddle.regularizer.L2Decay(0.0005), parameters=model.parameters())

    for epoch in range(args.epoch):
        for i, data in enumerate(train_loader()):
            img, gt_boxes, gt_labels, img_scale = data
            gt_scores = np.ones(gt_labels.shape).astype('float32')
            gt_scores = paddle.to_tensor(gt_scores)
            img = paddle.to_tensor(img)
            gt_boxes = paddle.to_tensor(gt_boxes)
            gt_labels = paddle.to_tensor(gt_labels)
            outputs = model(img)  #前向传播，输出[P0, P1, P2]
            loss = model.get_loss(outputs, gt_boxes, gt_labels, gtscore=gt_scores,
                                  anchors = ANCHORS,
                                  anchor_masks = ANCHOR_MASKS,
                                  ignore_thresh=IGNORE_THRESH,
                                  use_label_smooth=False)        # 计算损失函数

            loss.backward()    # 反向传播计算梯度
            opt.step()  # 更新参数
            opt.clear_grad()
            if i % 10 == 0:
                timestring = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(time.time()))
                print('{}[TRAIN]epoch {}, iter {}, output loss: {}'.format(timestring, epoch, i, loss.numpy()), flush=True)
        model.train()


    # save params of model
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir)

    paddle.save(model.state_dict(), args.output_dir + '/param.pdparams')



